==> stream-download-0.13.1.crate: stream-download-0.13.1/src/open_dal.rs <==
//! A [`SourceStream`] adapter for [OpenDAL](https://docs.rs/opendal/latest/opendal).
//! `OpenDAL` is a data access layer that supports data retrieval from a variety of storage
//! services. The list of supported services is [documented here](https://docs.rs/opendal/latest/opendal/services/index.html).
//!
//! # Example using S3
//!
//! ```no_run
//! use std::error::Error;
//!
//! use opendal::{Operator, services};
//! use stream_download::open_dal::{OpenDalStream, OpenDalStreamParams};
//! use stream_download::storage::temp::TempStorageProvider;
//! use stream_download::{Settings, StreamDownload};
//!
//! #[tokio::main]
//! async fn main() -> Result<(), Box<dyn Error>> {
//!     let mut builder = services::S3::default()
//!         .region("us-east-1")
//!         .access_key_id("test")
//!         .secret_access_key("test")
//!         .bucket("my-bucket");
//!
//!     let operator = Operator::new(builder)?.finish();
//!     let stream =
//!         OpenDalStream::new(OpenDalStreamParams::new(operator, "some-object-key")).await?;
//!
//!     Ok(())
//! }
//! ```

use std::fmt::Debug;
use std::io::{self};
use std::num::NonZeroUsize;
use std::task::Poll;

use bytes::{Bytes, BytesMut};
use futures::{Stream, ready};
use opendal::{FuturesAsyncReader, Metakey, Operator, Reader};
use pin_project_lite::pin_project;
use tokio_util::compat::{Compat, FuturesAsyncReadCompatExt};
use tokio_util::io::poll_read_buf;
use tracing::instrument;

use crate::WrapIoResult;
use crate::source::{DecodeError, SourceStream};

/// Parameters for creating an `OpenDAL` stream.
#[derive(Debug, Clone)]
pub struct OpenDalStreamParams {
    operator: Operator,
    path: String,
    chunk_size: usize,
}

impl OpenDalStreamParams {
    /// Creates a new [`OpenDalStreamParams`] instance.
    pub fn new<S>(operator: Operator, path: S) -> Self
    where
        S: Into<String>,
    {
        Self {
            operator,
            path: path.into(),
            chunk_size: 4096,
        }
    }

    /// Sets the chunk size for the [`OpenDalStream`].
    /// The default value is 4096.
    #[must_use]
    pub fn chunk_size(mut self, chunk_size: NonZeroUsize) -> Self {
        self.chunk_size = chunk_size.get();
        self
    }
}

pin_project! {
    /// An `OpenDAL` implementation of the [`SourceStream`] trait
    pub struct OpenDalStream {
        #[pin]
        async_reader: Compat<FuturesAsyncReader>,
        reader: Reader,
        buf: BytesMut,
        capacity: usize,
        content_length: Option<u64>,
        content_type: Option<String>,
    }
}

// Can't use educe here because of https://github.com/taiki-e/pin-project-lite/issues/3
impl Debug for OpenDalStream {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("OpenDalStream")
            .field("async_reader", &"<async_reader>")
            .field("reader", &"<reader>")
            .field("buf", &self.buf)
            .field("capacity", &self.capacity)
            .field("content_length", &self.content_length)
            .field("content_type", &self.content_type)
            .finish()
    }
}

/// Error returned from `OpenDAL`
#[derive(thiserror::Error, Debug)]
#[error("{0}")]
pub struct Error(#[from] opendal::Error);

impl DecodeError for Error {}

impl OpenDalStream {
    /// Creates a new [`OpenDalStream`].
    #[instrument]
    pub async fn new(params: OpenDalStreamParams) -> Result<Self, Error> {
        let stat = params.operator.stat(&params.path).await?;

        let content_length = if stat.metakey().contains(Metakey::ContentLength) {
            // content_length() will panic if called when the ContentLength Metakey is not present
            Some(stat.content_length())
        } else {
            None
        };

        let content_type = stat.content_type().map(|t| t.to_string());

        let reader = params.operator.reader(&params.path).await?;

        let async_reader = reader.clone().into_futures_async_read(..).await?.compat();

        Ok(Self {
            async_reader,
            reader,
            buf: BytesMut::with_capacity(params.chunk_size),
            capacity: params.chunk_size,
            content_length,
            content_type,
        })
    }

    /// Returns the content type of the stream, if it is known.
    pub fn content_type(&self) -> Option<&str> {
        self.content_type.as_deref()
    }
}

impl SourceStream for OpenDalStream {
    type Params = OpenDalStreamParams;

    type StreamCreationError = Error;

    async fn create(params: Self::Params) -> Result<Self, Self::StreamCreationError> {
        Self::new(params).await
    }

    fn content_length(&self) -> Option<u64> {
        self.content_length
    }

    async fn seek_range(&mut self, start: u64, end: Option<u64>) -> io::Result<()> {
        let reader = self.reader.clone();
        let async_reader = match end {
            Some(end) => reader.into_futures_async_read(start..end).await,
            None => reader.into_futures_async_read(start..).await,
        };

        self.async_reader = async_reader
            .map_err(|e| e.into())
            .wrap_err("error creating async reader")?
            .compat();
        Ok(())
    }

    async fn reconnect(&mut self, current_position: u64) -> io::Result<()> {
        self.seek_range(current_position, None).await
    }

    fn supports_seek(&self) -> bool {
        true
    }
}

impl Stream for OpenDalStream {
    type Item = io::Result<Bytes>;

    fn poll_next(
        mut self: std::pin::Pin<&mut Self>,
        cx: &mut std::task::Context<'_>,
    ) -> std::task::Poll<Option<Self::Item>> {
        let mut this = self.as_mut().project();

        if this.buf.capacity() == 0 {
            this.buf.reserve(*this.capacity);
        }

        match ready!(poll_read_buf(this.async_reader, cx, &mut this.buf)) {
            Err(err) => Poll::Ready(Some(Err(err))),
            Ok(0) => Poll::Ready(None),
            Ok(_) => {
                let chunk = this.buf.split();
                Poll::Ready(Some(Ok(chunk.freeze())))
            }
        }
    }
}

==> stream-download-0.14.1.crate: stream-download-0.14.1/src/open_dal.rs <==
//! A [`SourceStream`] adapter for [OpenDAL](https://docs.rs/opendal/latest/opendal).
//! `OpenDAL` is a data access layer that supports data retrieval from a variety of storage
//! services. The list of supported services is [documented here](https://docs.rs/opendal/latest/opendal/services/index.html).
//!
//! # Example using S3
//!
//! ```no_run
//! use std::error::Error;
//!
//! use opendal::{Operator, services};
//! use stream_download::open_dal::{OpenDalStream, OpenDalStreamParams};
//! use stream_download::storage::temp::TempStorageProvider;
//! use stream_download::{Settings, StreamDownload};
//!
//! #[tokio::main]
//! async fn main() -> Result<(), Box<dyn Error>> {
//!     let mut builder = services::S3::default()
//!         .region("us-east-1")
//!         .access_key_id("test")
//!         .secret_access_key("test")
//!         .bucket("my-bucket");
//!
//!     let operator = Operator::new(builder)?.finish();
//!     let stream =
//!         OpenDalStream::new(OpenDalStreamParams::new(operator, "some-object-key")).await?;
//!
//!     Ok(())
//! }
//! ```

use std::fmt::Debug;
use std::io::{self};
use std::num::NonZeroUsize;
use std::task::Poll;

use bytes::{Bytes, BytesMut};
use futures::{Stream, ready};
use opendal::{FuturesAsyncReader, Operator, Reader};
use pin_project_lite::pin_project;
use tokio_util::compat::{Compat, FuturesAsyncReadCompatExt};
use tokio_util::io::poll_read_buf;
use tracing::instrument;

use crate::WrapIoResult;
use crate::source::{DecodeError, SourceStream};

/// Parameters for creating an `OpenDAL` stream.
#[derive(Debug, Clone)]
pub struct OpenDalStreamParams {
    operator: Operator,
    path: String,
    chunk_size: usize,
}

impl OpenDalStreamParams {
    /// Creates a new [`OpenDalStreamParams`] instance.
    pub fn new<S>(operator: Operator, path: S) -> Self
    where
        S: Into<String>,
    {
        Self {
            operator,
            path: path.into(),
            chunk_size: 4096,
        }
    }

    /// Sets the chunk size for the [`OpenDalStream`].
    /// The default value is 4096.
    #[must_use]
    pub fn chunk_size(mut self, chunk_size: NonZeroUsize) -> Self {
        self.chunk_size = chunk_size.get();
        self
    }
}

pin_project! {
    /// An `OpenDAL` implementation of the [`SourceStream`] trait
    pub struct OpenDalStream {
        #[pin]
        async_reader: Compat<FuturesAsyncReader>,
        reader: Reader,
        buf: BytesMut,
        capacity: usize,
        content_length: Option<u64>,
        content_type: Option<String>,
    }
}

// Can't use educe here because of https://github.com/taiki-e/pin-project-lite/issues/3
impl Debug for OpenDalStream {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("OpenDalStream")
            .field("async_reader", &"<async_reader>")
            .field("reader", &"<reader>")
            .field("buf", &self.buf)
            .field("capacity", &self.capacity)
            .field("content_length", &self.content_length)
            .field("content_type", &self.content_type)
            .finish()
    }
}

/// Error returned from `OpenDAL`
#[derive(thiserror::Error, Debug)]
#[error("{0}")]
pub struct Error(#[from] opendal::Error);

impl DecodeError for Error {}

impl OpenDalStream {
    /// Creates a new [`OpenDalStream`].
    #[instrument]
    pub async fn new(params: OpenDalStreamParams) -> Result<Self, Error> {
        let stat = params.operator.stat(&params.path).await?;

        let content_length = stat.content_length();
        let content_type = stat.content_type().map(|t| t.to_string());

        let reader = params.operator.reader(&params.path).await?;

        let async_reader = reader.clone().into_futures_async_read(..).await?.compat();

        Ok(Self {
            async_reader,
            reader,
            buf: BytesMut::with_capacity(params.chunk_size),
            capacity: params.chunk_size,
            content_length: if content_length > 0 {
                Some(content_length)
            } else {
                None
            },
            content_type,
        })
    }

    /// Returns the content type of the stream, if it is known.
    pub fn content_type(&self) -> Option<&str> {
        self.content_type.as_deref()
    }
}

impl SourceStream for OpenDalStream {
    type Params = OpenDalStreamParams;

    type StreamCreationError = Error;

    async fn create(params: Self::Params) -> Result<Self, Self::StreamCreationError> {
        Self::new(params).await
    }

    fn content_length(&self) -> Option<u64> {
        self.content_length
    }

    async fn seek_range(&mut self, start: u64, end: Option<u64>) -> io::Result<()> {
        let reader = self.reader.clone();
        let async_reader = match end {
            Some(end) => reader.into_futures_async_read(start..end).await,
            None => reader.into_futures_async_read(start..).await,
        };

        self.async_reader = async_reader
            .map_err(|e| e.into())
            .wrap_err("error creating async reader")?
            .compat();
        Ok(())
    }

    async fn reconnect(&mut self, current_position: u64) -> io::Result<()> {
        self.seek_range(current_position, None).await
    }

    fn supports_seek(&self) -> bool {
        true
    }
}

impl Stream for OpenDalStream {
    type Item = io::Result<Bytes>;

    fn poll_next(
        mut self: std::pin::Pin<&mut Self>,
        cx: &mut std::task::Context<'_>,
    ) -> std::task::Poll<Option<Self::Item>> {
        let mut this = self.as_mut().project();

        if this.buf.capacity() == 0 {
            this.buf.reserve(*this.capacity);
        }

        match ready!(poll_read_buf(this.async_reader, cx, &mut this.buf)) {
            Err(err) => Poll::Ready(Some(Err(err))),
            Ok(0) => Poll::Ready(None),
            Ok(_) => {
                let chunk = this.buf.split();
                Poll::Ready(Some(Ok(chunk.freeze())))
            }
        }
    }
}
